import logging
from collections import OrderedDict


def _is_within(parent, o, ignore: set):
    """
    Determine of an object is within a parent object.
    :param parent: The object in which o hopefully exists.
    :param o: The object to find in parent.
    :param ignore: A set of keys to ignore in parent.
    return: True if o is within parent somewhere. False otherwise.
    """
    if not isinstance(parent, type(o)):
        return False
    elif isinstance(o, dict):
        return _is_within_dict(parent, o, ignore)
    else:
        return parent == o


def _is_within_dict(parent, o, ignore: set):
    """
    Determine of an object is within a parent dict object.
    :param parent: The object in which o hopefully exists.
    :param o: The object to find in parent.
    :param ignore: A set of keys to ignore in parent.
    return: True if o is within parent somewhere. False otherwise.
    """
    ret = False
    for k, v in o.items():
        if k in ignore:
            ret = True
            break
        elif k in parent and parent[k] == v:
            ret = True
            break
    return ret


def get_from_list(hub, parent, o):
    """
    Returns the first object found in parent that contains o, None if no object is found.
    :param hub: The redistributed pop central hub. This is required in
    Idem, so while not used, must appear.
    :param parent: An iterable object to search for o.
    :param o: An object the values of which must exist in one of the iterables.

    For example:

    The subset:

    { name: "my_object" }

    exists within (and would be returned)

    [
        { name: "not_my_object", something_else: "not the other thing" }
        { name: "my_object", something_else: "some other thing"},
    ]
    """
    ret = None

    for item in parent:
        if isinstance(item, dict):
            if _is_within(item, o, set()):
                ret = item
                break
        else:
            logging.warning(f"Item {str(item)} should be a dict. Skipping this item.")

    return ret


def _no_update(old_value, new_value) -> bool:
    """Given two value, check if there is a need to update from the old_value to new_value."""
    if isinstance(old_value, str) and isinstance(new_value, str):
        # If the two values are the same string, then no need to update.
        return old_value == new_value
    elif isinstance(old_value, dict) and isinstance(new_value, dict):
        # If the new value dictionary is a subset of the old value dictionary, then no need to update
        return new_value.items() <= old_value.items()
    else:
        return False


def patch_json_content(
    hub,
    patch_parameters: dict,
    old_values: dict,
    new_values: dict,
    force_update_parameters: dict = {"tags": "tags"},
) -> dict:
    """
    Generate a json that contains all the parameters and their values that will be sent during a PATCH operation
    :param hub: The redistributed pop central hub. This is required in
    Idem, so while not used, must appear.
    :param patch_parameters: A dictionary of patchable parameters.
    For example: {"tags": "tags", "properties": {"properties_1": "properties_1", "properties_2": "properties_2"}}
    :param old_values: A dictionary that contains the old values of parameters.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    :param new_values: A dictionary that contains the new values of parameters. This should be the exact structure as
     what old_values have.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    :param force_update_parameters: A dictionary that determines what parameters will be updated as long as the value is
    different, regardless if the new value is a sub-set of old value or not.
    For example: {"tags": "new-tag", "properties": {"properties_1": "value_1", "properties_2": "value_2"}}
    """
    payload = {}
    for parameter_key, parameter_fields in patch_parameters.items():
        value = new_values.get(parameter_key, None)
        if value is None:
            continue
        elif isinstance(parameter_fields, str):
            if parameter_fields in old_values:
                # If the new value is a sub-set of the old value, then we need to decide if this should be a patch
                # operation or a no-op. For example, if tags property changes from {tagA, tagB} to {tagA}, this should
                # trigger a patch operation to update tags property to remove tagB. However, for some properties,
                # seeing a property being absent doesn't mean that we want to delete that property. So to distinguish
                # these two scenarios, we use 'force_update_parameters' to decide what should be updated as long as
                # the value is different.
                if parameter_fields in force_update_parameters:
                    # If a property is in 'force_update_parameters', update the resource as long as old and new value
                    # are different
                    if old_values[parameter_fields] != value:
                        payload.update({parameter_key: value})
                else:
                    # If a property is not in 'force_update_parameters', update the resource only if old and
                    # new value are different and new value is not a sub-set of the old value
                    if old_values[parameter_fields] != value and (
                        not _no_update(old_values[parameter_fields], value)
                    ):
                        payload.update({parameter_key: value})
            else:
                payload.update({parameter_key: value})
        elif isinstance(parameter_fields, dict):
            sub_force_update = force_update_parameters.get(parameter_key, dict())
            if parameter_key in old_values:
                sub_payload = patch_json_content(
                    hub,
                    parameter_fields,
                    old_values[parameter_key],
                    value,
                    sub_force_update,
                )
            else:
                sub_payload = patch_json_content(
                    hub, parameter_fields, dict(), value, sub_force_update
                )
            if sub_payload:
                payload.update({parameter_key: sub_payload})
        else:
            continue
    return payload


def get_uri_parameter_value_from_uri(hub, uri: str, parameters: OrderedDict) -> list:
    """
    Generate a list of pairs of uri parameter and its value. And convert uri parameters to expected key used in sls files
    For example /virtualNetworks/virtual-network-name/subnets/subnet-name} will output
    [{"virtualNetworks": "virtual-network-name"}, {"subnets": "subnet-name"}]
    :param hub: The redistributed pop central hub. This is required in Idem, so while not used, must appear.
    :param uri: the uri
    :param parameters: parameters to search through the uri.
    For example: OrderedDict({"virtualNetworks": "virtual_network_name"})
    """
    ret = []
    uri_values = uri.split("/")
    for parameter_key, parameter_value in parameters.items():
        if parameter_key in uri_values:
            uri_index = uri_values.index(parameter_key)
            # Convert parameters from camel case to snake case
            ret.append({parameter_value: uri_values[uri_index + 1]})
    return ret



import os


try:
    import bs4
    import requests
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


__func_alias__ = {"type_": "type"}


def parse(hub, api_url: str):
    plugins = {}

    with requests.get(f"{api_url}/en-us/rest/api") as response:
        html = response.text
    soup = bs4.BeautifulSoup(html, "lxml")

    resources = []
    num = 1
    while num:
        q = soup.find("meta", {"name": f"quickFilterColumn{num}"})

        if q:
            r = q.attrs["content"].split(",")
            resources.extend(r)
        else:
            break
        num += 1

    for r in tqdm.tqdm(resources):
        ref = f"{r}"
        if ref.lower() == "azure":
            file_path = os.path.dirname(__file__) + "/resource/azure-rest-api.html"
        else:
            hub.log.error(f"{ref} is not a supported resource.")
            continue
        # url = f"https://docs.microsoft.com/en-us/rest/api/?view={r}"
        new_plugins = hub.pop_create.azure.api_reference.parse(file_path)

        for k, v in new_plugins.items():
            ref = f"{hub.tool.format.case.snake(r)}.{hub.tool.format.case.snake(k)}"
            plugins[ref] = v

    return plugins



from typing import Any
from typing import Dict
from typing import List


async def paginate(
    hub, ctx, url: str, success_codes: List = [200], headers: Dict = {}
) -> Dict[str, Any]:
    """
    Paginate items from the given azure url
    :param hub: The redistributed pop central hub.
    :param ctx: A dict with the keys/values for the execution of the Idem run located in
     `hub.idem.RUNS[ctx['run_name']]`.
    :param url: HTTP request url
    :param success_codes: List of HTTP status codes that are considered as "successful" to a request
    :param headers: HTTP request headers
    :return: Response body
    """
    while url:
        ret = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=success_codes,
            headers=headers,
        )
        if not ret["result"]:
            raise ValueError(
                f"Error on requesting GET {url} with status code {ret['status']}:"
                f" {ret.get('comment', '')}"
            )
        result = ret["ret"]
        yield result
        url = result.get("nextLink", None)



import os
import pathlib
import traceback
from os.path import exists

import tqdm
from dict_tools.data import NamespaceDict

try:
    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)

    ctx.servers = [
        "https://management.azure.com/",
        "https://management.core.windows.net/",
    ]

    # We already have an acct plugin
    ctx.has_acct_plugin = False
    ctx.service_name = ctx.service_name or "azure_auto"
    docs_api = "https://docs.microsoft.com"

    # plugins = hub.pop_create.azure.resource.parse(docs_api)

    # TODO: This list below will be replaced once the webpages parsers getting finished
    resources = [
        {
            "service": "Resource Management",
            "resource": "Resource Groups",
            "create": "https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/create-or-update",
            "update": "https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/update",
            "list": "https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/list",
            "delete": "https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/delete",
            "get": "https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/get",
        },
        {
            "service": "Virtual Networks",
            "resource": "Virtual Networks",
            "create": "https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/create-or-update",
            "update": "https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/update-tags",
            "list": "https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/list-all",
            "delete": "https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/delete",
            "get": "https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/get",
        },
    ]

    # Initialize cloud spec
    ctx.cloud_spec = NamespaceDict(
        project_name=ctx.project_name,
        service_name=ctx.service_name,
        request_format={
            "present": hub.pop_create.azure.template.PRESENT_REQUEST_FORMAT,
            "absent": hub.pop_create.azure.template.ABSENT_REQUEST_FORMAT,
            "describe": hub.pop_create.azure.template.DESCRIBE_REQUEST_FORMAT,
        },
        plugins={},
    )

    for resource in tqdm.tqdm(resources):
        try:
            plugins = plugin(hub, resource)
            if plugins:
                ctx.cloud_spec.plugins = plugins
                hub.cloudspec.init.run(
                    ctx,
                    directory,
                    create_plugins=["state_modules"],
                )
        except Exception as e:
            hub.log.error(
                f"Error when generating {resource.get('resource')} with error {traceback.print_stack(e)}"
            )

    hub.pop_create.init.run(
        directory=directory,
        subparsers=["cicd"],
        **ctx,
    )

    ctx.cloud_spec.plugins = {}
    return ctx


def plugin(hub, resource):
    service_name_formatted = resource.get("service").lower().strip().replace(" ", "_")
    resource_name_formatted = resource.get("resource").lower().strip().replace(" ", "_")
    plugin_key = f"{service_name_formatted}.{resource_name_formatted}"
    plugins = dict()
    if exists(
        f"{os.path.dirname(os.path.realpath(__file__))}/../../states/azure_auto/{service_name_formatted}/{resource_name_formatted}.py"
    ):
        # if the auto-generated file has existed, skip generating this resource
        return plugins
    shared_function_data = {
        "get_function": f"await hub.exec.request.json.get",
        "create_function": f"await hub.exec.request.json.put",
        "list_function": f"await hub.exec.request.json.get",
        "delete_function": f"await hub.exec.request.raw.delete",
        "patch_function": f"await hub.exec.request.json.patch",
        "resource_name": resource.get("resource"),
        "create_ref": plugin_key,
    }
    resource_plugin = hub.pop_create.azure.plugin.parse(resource, shared_function_data)
    plugins[plugin_key] = resource_plugin
    return plugins



class DescribeError(Exception):
    """Exception raised for resources not found in Azure."""

    def __init__(self, message: str):
        """
        Initialize an instance of this class.
        :param message: String message to present when rendering the exception.
        """
        self.message = message
        super().__init__(self.message)



NAME_PARAMETER = {
    "default": None,
    "doc": "The identifier for this state",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}


PRESENT_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id
    response_get = {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200],
        headers=ctx.acct.headers,
    )

    if force_update:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would force to update azure.{{ function.hardcoded.create_ref }}",
            )
        response_force_put = {{ function.hardcoded.create_function }}(
                    ctx,
                    url=f"{{ function.hardcoded.path }}",
                    success_codes=[200, 201],
                    headers=ctx.acct.headers,
                    json=parameters,
        )
        if response_force_put["result"]:
            old_resource = response_get["ret"] if response_get["result"] else None
            return StateReturn(
                name=name,
                result=True,
                old_obj=old_resource,
                new_obj=response_force_put["ret"],
                comment=response_force_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not force to update {{ function.hardcoded.resource_name }} {response_force_put['comment']} {response_force_put['ret']}"
            )
            return StateReturn(
                name=name,
                result=False,
                comment=response_force_put["comment"],
                error=response_force_put["ret"],
            )

    if not response_get["result"]:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would create azure.{{ function.hardcoded.create_ref }}"
            )

        if response_get["status"] == 404:
            # PUT operation to create a resource
            response_put = {{ function.hardcoded.create_function }}(
                ctx,
                url=f"{{ function.hardcoded.path }}",
                success_codes=[200, 201],
                headers=ctx.acct.headers,
                json=parameters,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create {{ function.hardcoded.resource_name }} {response_put['comment']} {response_put['ret']}"
                )
                return StateReturn(
                    name=name, result=False, comment=response_put["comment"], error=response_put['ret']
                )

            return StateReturn(
                name=name,
                result=True,
                old_obj=None,
                new_obj=response_put["ret"],
                comment=response_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not get {{ function.hardcoded.resource_name }} {response_get['comment']} {response_get['ret']}"
            )
            return StateReturn(
                name=name, result=False, comment=response_get["comment"], error=response_get['ret']
            )
    {% if function.hardcoded.patch_parameters %}
    else:
        # PATCH operation to update a resource
        patch_parameters = {{ function.hardcoded.patch_parameters }}
        existing_resource = response_get["ret"]
        new_parameters = hub.tool.azure.utils.patch_json_content(patch_parameters, existing_resource, parameters)
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment=f"Would update azure.{{ function.hardcoded.create_ref }} with parameters: {new_parameters}"
            )

        if not new_parameters:
            return StateReturn(
                name=name,
                result=True,
                old_obj=existing_resource,
                new_obj=existing_resource,
                comment=f"'{name}' has no property need to be updated.",
            )

        response_patch = {{ function.hardcoded.patch_function }}(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200],
            headers=ctx.acct.headers,
            json=new_parameters,
        )

        if not response_patch["result"]:
            hub.log.debug(
                f"Could not update {{ function.hardcoded.resource_name }} {response_patch['comment']} {response_patch['ret']}"
            )
            return StateReturn(
                name=name, result=False, comment=response_patch["comment"], error=response_patch['ret']
            )

        return StateReturn(
            name=name,
            result=True,
            old_obj=existing_resource,
            new_obj=response_patch["ret"],
            comment=response_patch["comment"],
        )
    {% else %}
    # No update operation on {{ function.hardcoded.resource_name }} since Azure does not have PATCH api on {{ function.hardcoded.resource_name }}
    return StateReturn(
        name=name,
        result=True,
        old_obj=response_get["ret"],
        new_obj=response_get["ret"],
        comment=response_get["comment"],
    )
    {% endif %}
"""

ABSENT_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id
    response_get = {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200],
        headers=ctx.acct.headers,
    )
    if response_get["result"]:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would delete azure.{{ function.hardcoded.create_ref }}"
            )

        existing_resource = response_get["ret"]
        response_delete = {{ function.hardcoded.delete_function }}(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200, 202, 204],
            headers=ctx.acct.headers,
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete {{ function.hardcoded.resource_name }} {response_delete['comment']} {response_delete['ret']}"
            )
            return StateReturn(
                name=name, result=False, comment=response_delete["comment"], error=response_delete['ret']
            )

        return StateReturn(
            name=name,
            result=True,
            old_obj=existing_resource,
            new_obj={},
            comment=response_delete["comment"],
        )
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        return StateReturn(
            name=name,
            result=True,
            old_obj=None,
            new_obj=None,
            comment=f"'{name}' already absent",
        )
    else:
        hub.log.debug(
            f"Could not get {{ function.hardcoded.resource_name }} {response_get['comment']} {response_get['ret']}"
        )
        return StateReturn(
            name=name, result=False, comment=response_get["comment"], error=response_get['ret']
        )
"""

DESCRIBE_REQUEST_FORMAT = r"""
    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict({{ function.hardcoded.describe_parameters }})
    try:
        async for page_result in hub.tool.azure.request.paginate(
                ctx,
                url=f"{{ function.hardcoded.path }}",
                success_codes=[200],
                headers=ctx.acct.headers,
        ):
            resource_list = page_result.get("value", None)
            if resource_list:
                for resource in resource_list:
                    uri_parameter_values = (
                        hub.tool.azure.utils.get_uri_parameter_value_from_uri(resource["id"], uri_parameters)
                    )
                    result[resource["id"]] = {f"azure.{{ function.hardcoded.create_ref }}.present": uri_parameter_values
                    + [{"parameters": resource}]}
    except ValueError as e:
        raise DescribeError(f"Error on describing {{ function.hardcoded.resource_name }}: {str(e)}")
    return result
"""



import logging

try:
    import bs4
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def parse(hub, file_path: str, base_url="https://docs.microsoft.com"):
    """
    Parse a page such as this one: https://docs.microsoft.com/en-us/rest/api/?view=Azure

    Create a pretty version of a list of Azure services"
    """
    plugins = {}
    with open(file_path) as f:
        html = f.read()
    soup = bs4.BeautifulSoup(html, "lxml")

    # Parse page and get links, names, and descriptions of each Azure service
    parsing_results = []
    api_table = soup.find(class_="api-search-results")
    table_body = api_table.find("tbody")
    for row in table_body.find_all("tr"):
        try:
            cols = row.find_all("td")
            parsing_result = {
                "name": cols[0].find("a").text,
                "url": base_url + cols[0].find("a")["href"],
                "description": cols[1].text,
            }
            parsing_results.append(parsing_result)
        except [AttributeError, KeyError] as parse_error:
            logging.error(
                f"Fail to parse table from file {file_path} at row {str(row)}. Error: {parse_error}"
            )

    for api_reference_data in tqdm.tqdm(parsing_results):
        plugins[api_reference_data["name"]] = hub.pop_create.azure.rest.parse(
            url=api_reference_data["url"], desc=api_reference_data["description"]
        )

    return plugins



import dict_tools.differ as differ


class StateReturn(dict):
    """
    Convenience class to manage POP state returns.
    """

    def __init__(
        self,
        name=None,
        result=None,
        comment=None,
        old_obj=None,
        new_obj=None,
        error=None,
    ):
        """
        Initialize an object of this class.
        :param name: The name of the state (e.g. from a SLS file).
        :param result: True if the state call works, False otherwise.
        :param comment: Any relevant comments to the state execution.
        For example a 200 code from an HTTP call or a mere readable comment.
        :param old_obj: For state changes, the object's state
        (e.g., dict of values).
        prior to any state change request executions.
        :param new_obj: For state changes, the new object's state after any
        state change request executions.
        """
        super().__init__(
            [
                ("name", name),
                ("result", result),
                ("comment", comment),
                ("changes", None),
            ]
        )
        if old_obj or new_obj:
            self["changes"] = differ.deep_diff(
                old_obj if old_obj else dict(), new_obj if new_obj else dict()
            )
        if error:
            comment += " " + str(error)
        if comment:
            self["comment"] = comment



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
from collections import OrderedDict
from typing import Any
from typing import Dict

from idem_azure_auto.helpers.exc import DescribeError
from idem_azure_auto.helpers.returns import StateReturn

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    policy_assignment_name: str,
    parameters: dict = {},
    force_update: bool = False,
) -> StateReturn:
    r"""
    **Autogenerated function**

    Create or update Policy Assignments

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name of the policy assignment.
        parameters(dict, optional): API request payload parameters. Defaults to {}.
        force_update(bool, optional): If PUT operation should be used instead of PATCH operation during resource update. Defaults to False.

    Returns:
        StateReturn

    Examples:

        .. code-block:: sls

            resource_is_present:
              azure.policy.policy_assignments.present:
                - name: value
                - scope: value
                - policy_assignment_name: value
    """

    subscription_id = ctx.acct.subscription_id
    if "{subscriptionId}" in scope:
        scope = scope.replace("{subscriptionId}", subscription_id)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
        success_codes=[200],
        headers=ctx.acct.headers,
    )

    if force_update:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would force to update azure.policy.policy_assignments",
            )
        response_force_put = await hub.exec.request.json.put(
            ctx,
            url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
            success_codes=[201],
            headers=ctx.acct.headers,
            json=parameters,
        )
        if response_force_put["result"]:
            old_resource = response_get["ret"] if response_get["result"] else None
            return StateReturn(
                name=name,
                result=True,
                old_obj=old_resource,
                new_obj=response_force_put["ret"],
                comment=response_force_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not force to update Policy Assignments {response_force_put['comment']} {response_force_put['ret']}"
            )
            return StateReturn(
                name=name,
                result=False,
                comment=response_force_put["comment"],
                error=response_force_put["ret"],
            )

    if not response_get["result"]:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would create azure.policy.policy_assignments",
            )

        if response_get["status"] == 404:
            # PUT operation to create a resource
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
                success_codes=[201],
                headers=ctx.acct.headers,
                json=parameters,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create Policy Assignments {response_put['comment']} {response_put['ret']}"
                )
                return StateReturn(
                    name=name,
                    result=False,
                    comment=response_put["comment"],
                    error=response_put["ret"],
                )

            return StateReturn(
                name=name,
                result=True,
                old_obj=None,
                new_obj=response_put["ret"],
                comment=response_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not get Policy Assignments {response_get['comment']} {response_get['ret']}"
            )
            return StateReturn(
                name=name,
                result=False,
                comment=response_get["comment"],
                error=response_get["ret"],
            )

    else:
        # PATCH operation to update a resource
        patch_parameters = {"identity": "identity", "location": "location"}
        existing_resource = response_get["ret"]
        new_parameters = hub.tool.azure.utils.patch_json_content(
            patch_parameters, existing_resource, parameters, force_update_parameters={}
        )
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment=f"Would update azure.policy.policy_assignments with parameters: {new_parameters}",
            )

        if not new_parameters:
            return StateReturn(
                name=name,
                result=True,
                old_obj=existing_resource,
                new_obj=existing_resource,
                comment=f"'{name}' has no property need to be updated.",
            )

        response_patch = await hub.exec.request.json.patch(
            ctx,
            url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
            success_codes=[200],
            headers=ctx.acct.headers,
            json=new_parameters,
        )

        if not response_patch["result"]:
            hub.log.debug(
                f"Could not update Policy Assignments {response_patch['comment']} {response_patch['ret']}"
            )
            return StateReturn(
                name=name,
                result=False,
                comment=response_patch["comment"],
                error=response_patch["ret"],
            )

        return StateReturn(
            name=name,
            result=True,
            old_obj=existing_resource,
            new_obj=response_patch["ret"],
            comment=response_patch["comment"],
        )


async def absent(
    hub, ctx, name: str, scope: str, policy_assignment_name: str
) -> StateReturn:
    r"""
    **Autogenerated function**

    Delete Policy Assignments

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name of the policy assignment to delete.

    Returns:
        StateReturn

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.policy.policy_assignments.absent:
                - name: value
                - scope: value
                - policy_assignment_name: value
    """

    subscription_id = ctx.acct.subscription_id
    if "{subscriptionId}" in scope:
        scope = scope.replace("{subscriptionId}", subscription_id)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
        success_codes=[200],
        headers=ctx.acct.headers,
    )
    if response_get["result"]:
        if ctx.get("test", False):
            return StateReturn(
                name=name,
                result=True,
                comment="Would delete azure.policy.policy_assignments",
            )

        existing_resource = response_get["ret"]
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"https://management.azure.com/{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}?api-version=2021-06-01",
            success_codes=[200, 204],
            headers=ctx.acct.headers,
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete Policy Assignments {response_delete['comment']} {response_delete['ret']}"
            )
            return StateReturn(
                name=name,
                result=False,
                comment=response_delete["comment"],
                error=response_delete["ret"],
            )

        return StateReturn(
            name=name,
            result=True,
            old_obj=existing_resource,
            new_obj={},
            comment=response_delete["comment"],
        )
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        return StateReturn(
            name=name,
            result=True,
            old_obj=None,
            new_obj=None,
            comment=f"'{name}' already absent",
        )
    else:
        hub.log.debug(
            f"Could not get Policy Assignments {response_get['comment']} {response_get['ret']}"
        )
        return StateReturn(
            name=name,
            result=False,
            comment=response_get["comment"],
            error=response_get["ret"],
        )


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Policy Assignments under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.policy.policy_assignments
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict({"policyAssignments": "policy_assignment_name"})
    try:
        async for page_result in hub.tool.azure.request.paginate(
            ctx,
            url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Authorization/policyAssignments?api-version=2021-06-01",
            success_codes=[200],
            headers=ctx.acct.headers,
        ):
            resource_list = page_result.get("value", None)
            if resource_list:
                for resource in resource_list:
                    uri_parameter_values = (
                        hub.tool.azure.utils.get_uri_parameter_value_from_uri(
                            resource["id"], uri_parameters
                        )
                    )
                    if uri_parameter_values is None:
                        uri_parameter_values = list()
                    uri_parameter_values.insert(
                        0, {"scope": resource.get("properties").get("scope")}
                    )
                    result[resource["id"]] = {
                        f"azure.policy.policy_assignments.present": uri_parameter_values
                        + [{"parameters": resource}]
                    }
    except ValueError as e:
        raise DescribeError(f"Error on describing Policy Assignments: {str(e)}")
    return result



import asyncio
import sys


def __virtual__(hub):
    if "pytest" in sys.modules:
        return True
    else:
        return False, "Not running in test environment"


async def wait_for_resource_to_present(
    hub, ctx, url: str, retry_count: int, retry_period: int, retry_policy: list = [404]
):
    """
    Wait for a resource until its provisionState reaches "Succeed". This function is helpful to sequentially provision
    resources that depend on each other.
    :param hub: The redistributed pop central hub. This is required in Idem, so while not used, must appear.
    :param ctx: Idem ctx.
    :param url: The full url to get the resource from Azure.
    :param retry_count: The maximum number of retries to do.
    :param retry_period: The waiting time between each retry.
    :param retry_policy: A list of HTTP status code on which a retry should happen.
    """
    count = 0
    while count < retry_count:
        response_get = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=[200, 201, 204],
            headers=ctx.acct.headers,
        )
        if response_get["result"]:
            ret = response_get["ret"]
            if "properties" in ret and "provisioningState" in ret.get("properties"):
                if ret.get("properties").get("provisioningState") == "Succeeded":
                    return response_get
            else:
                count += 1
                hub.log.info(f"{count}/{retry_count} Wait for resource with url {url}")
                await asyncio.sleep(retry_period)
        elif response_get["status"] in retry_policy:
            count += 1
            hub.log.info(f"{count}/{retry_count} Wait for resource with url {url}")
            await asyncio.sleep(retry_period)
        else:
            raise RuntimeError(
                f"Getting resource with url {url} failed with status {response_get['status']}"
                f" and error: {response_get['ret']}"
            )
    raise RuntimeError(
        f"Resource with url {url} did not reach 'Succeeded' state within time limit."
    )


async def wait_for_resource_to_absent(
    hub,
    ctx,
    url: str,
    retry_count: int,
    retry_period: int,
    retry_policy: list = [200, 201, 204],
):
    """
    Wait for a resource until it's absent (GET operation returns 404).
    :param hub: The redistributed pop central hub. This is required in Idem, so while not used, must appear.
    :param ctx: Idem ctx.
    :param url: The full url to get the resource from Azure.
    :param retry_count: The maximum number of retries to do.
    :param retry_period: The waiting time between each retry.
    :param retry_policy: A list of HTTP status code on which a retry should happen.
    """
    count = 0
    while count < retry_count:
        response_get = await hub.exec.request.json.get(
            ctx,
            url=url,
            success_codes=[200, 201, 204],
            headers=ctx.acct.headers,
        )
        if response_get["result"]:
            count += 1
            hub.log.info(
                f"{count}/{retry_count} Wait for resource with url {url} to be deleted"
            )
            await asyncio.sleep(retry_period)
        elif response_get["status"] in retry_policy:
            count += 1
            hub.log.info(
                f"{count}/{retry_count} Wait for resource with url {url} to be deleted"
            )
            await asyncio.sleep(retry_period)
        elif response_get["status"] == 404:
            return
        else:
            raise RuntimeError(
                f"Getting resource with url {url} failed with status {response_get['status']}"
                f" and error: {response_get['ret']}"
            )
    raise RuntimeError(
        f"Resource with url {url} did not get deleted within time limit."
    )


def check_response_payload(hub, expected_payload: dict, actual_payload: dict):
    """
    Check if a response payload machines the expected payload.
    """
    assert expected_payload is not None
    assert actual_payload is not None
    for parameter_key, parameter_fields in expected_payload.items():
        assert (
            parameter_key in actual_payload
        ), f"{parameter_key} is not in {actual_payload}"
        if isinstance(parameter_fields, str):
            assert parameter_fields == actual_payload.get(
                parameter_key
            ), f"{parameter_fields} is not {actual_payload.get(parameter_key)}"
        elif isinstance(parameter_fields, list):
            for parameter_field in parameter_fields:
                if isinstance(parameter_field, str):
                    assert parameter_field in actual_payload.get(
                        parameter_key
                    ), f"{parameter_fields} is not in {actual_payload.get(parameter_key)}"
                else:
                    # Only support comparing a list of strings.
                    continue
        elif isinstance(parameter_fields, dict):
            check_response_payload(
                hub, parameter_fields, actual_payload.get(parameter_key)
            )
        else:
            # Skip any other type of fields that this function doesn't support
            continue



DYNE = {
    "pop_create": ["autogen"],
    "tool": ["tool"],
    "states": ["states"],
    "exec": ["exec"],
    "acct": ["acct"],
}
SUBCOMMANDS = {
    "azure": {
        "help": "Create idem_azure state modules by Azure REST API web",
        "dyne": "pop_create",
    },
}
